@@ -9,10 +9,12 @@ import importlib
 import itertools
 import numpy as np
 import os
+import pathlib
 import re
 import sympy
 import sys
 import traceback
+import typing
 import vorpy.tensor
 
 def __is_python_identifier (s):
@@ -22,6 +24,8 @@ def __is_python_identifier (s):
     return len(s) > 0 and s[0] in alphabetical_or_underscore and all(c in alphabetical_or_underscore or c in digits for c in s[1:])
 
 def __one_pass_replacement (s, replacement_d):
+    if len(replacement_d) == 0:
+        return s
     # This was taken from http://stackoverflow.com/questions/6116978/python-replace-multiple-strings
     escaped_replacement_d = dict((re.escape(k),v) for k,v in replacement_d.items())
     pattern = re.compile('|'.join(escaped_replacement_d.keys()))
@@ -94,20 +98,20 @@ def differential (F, X):
     if m == 0:
         # Univariate derivative
         if n == 0:
-            return F.diff(X)
+            return sympy.diff(F, X)
         # Multivariate derivative
         else:
-            return np.array([F.diff(X[I]) for I in multiindex_iterator(np.shape(X))]).reshape(np.shape(X))
+            return np.array([sympy.diff(F, X[I]) for I in multiindex_iterator(np.shape(X))]).reshape(np.shape(X))
 
     # Multivariate function
     else:
         # Univariate derivative
         if n == 0:
-            return np.array([F[I].diff(X) for I in multiindex_iterator(np.shape(F))]).reshape(np.shape(F))
+            return np.array([sympy.diff(F[I], X) for I in multiindex_iterator(np.shape(F))]).reshape(np.shape(F))
         # Multivariate derivative
         else:
             retval_shape = tuple(list(np.shape(F))+list(np.shape(X)))
-            return np.array([F[I[:m]].diff(X[I[m:]]) for I in multiindex_iterator(retval_shape)]).reshape(retval_shape)
+            return np.array([sympy.diff(F[I[:m]], X[I[m:]]) for I in multiindex_iterator(retval_shape)]).reshape(retval_shape)
 
 def D (F, *X_v):
     """
@@ -130,10 +134,10 @@ def D (F, *X_v):
 
     i.e. the Hessian of F.
     """
-    compiled_function = F
+    retval = F
     for X in X_v:
-        compiled_function = differential(compiled_function, X)
-    return compiled_function
+        retval = differential(retval, X)
+    return retval
 
 def python_expression_for (F, X, *, replacement_d={}, argument_id='X'):
     """
@@ -266,7 +270,7 @@ def cached_lambdified (function_id, *, function_creator, cache_dirname='lambdifi
     # TODO: make force_recache option
 
     assert __is_python_identifier(function_id)
-    assert __is_python_identifier(cache_dirname)
+    assert __is_python_identifier(cache_dirname) # TODO: Allow nested dirs here
 
     function_module_name = cache_dirname + '.' + function_id
 
@@ -318,6 +322,41 @@ def cached_lambdified (function_id, *, function_creator, cache_dirname='lambdifi
 
     return function_module.__dict__[function_id]
 
+def cache_lambdify (*, function_id:str, argument_id:str, replacement_d:typing.Dict[str,str]={}, import_v:typing.List[str]=[], decorator_v:typing.List[str]=[], cache_dirname_p:pathlib.Path=pathlib.Path('lambdified_cache'), verbose:bool=False):
+    """
+    This is a decorator function which makes it easy to obtain cached_lambdified functions.  Example:
+
+        import numpy as np
+        import sympy as sp
+        import vorpy.symbolic
+
+        def F_symbolic (Q):
+            x, y, z = Q
+            return x**2 - y*sp.sin(z) + sp.sqrt(x+y+z)
+
+        @vorpy.symbolic.cache_lambdify(
+            function_id='F',
+            argument_id='Q',
+            replacement_d={'dtype=object':'dtype=float', 'sqrt':'np.sqrt', 'sin':'np.sin'},
+            import_v=['import numpy as np'],
+            verbose=True,
+        )
+        def F_fast ():
+            x, y, z = sp.var('x,y,z')
+            Q = np.array([x,y,z])
+            return F_symbolic(Q), Q
+
+    Reference: https://python-3-patterns-idioms-test.readthedocs.io/en/latest/PythonDecorators.html
+    """
+    def wrap (generate_f_and_arguments):
+        def function_creator ():
+            f, arguments = generate_f_and_arguments()
+            return f, arguments, replacement_d, argument_id, import_v, decorator_v
+
+        return vorpy.symbolic.cached_lambdified(function_id, function_creator=function_creator, cache_dirname=str(cache_dirname_p), verbose=verbose)
+
+    return wrap
+
 def homogeneous_polynomial (coefficient_prefix, degree, X):
     """
     Returns a generic homogeneous polynomial of the given degree with symbolic coefficients,
